from ._GMM import GMM
from ._KDE import KDE

def VAE(device: 'str | device' = 'auto',
        layers_sizes: 'list | None' = None, 
        latent_dimension: 'int | None' = None, layers: 'int | None' = None,
        epochs: 'int | None' = None, search_trials: 'int' = 10,
        batches: 'int | None' = None, batch_size: 'int | None' = 128,
        learning_rate: 'float' = 1e-1,
        reduce_learning_rate: 'bool' = True, min_learning_rate: 'float' = 1e-5,
        patience: 'int | None' = None, time_limit: 'float' = 10):
    # Avoid importing more packages than necessary when possible
    from ._VAE import VAE as _VAE
    return _VAE(device=device,
            layers_sizes=layers_sizes, 
            latent_dimension=latent_dimension,
            layers=layers,
            epochs=epochs,
            search_trials=search_trials,
            batches=batches, batch_size=batch_size,
            learning_rate=learning_rate,
            reduce_learning_rate=reduce_learning_rate,
            min_learning_rate=min_learning_rate,
            patience=patience,
            time_limit=time_limit)

__all__ = ['GMM', 'KDE', 'VAE']


# -*- coding: utf-8 -*-
"""
Kernel Density Estimation

@author: david
"""

import numpy as np

from ._base import Generator, if_fitted, fit_method

def constant_h(X):
    n, dim = X.shape
    variance = np.power(n - 1, -1 / dim)
    return np.sqrt(variance)

def optimize_h(X):
    n, dim = X.shape
    separator = int(n * 0.6)
    ind = np.random.choice(n, n, False)
    X_train = X[ind[:separator]]
    X_validation = X[ind[separator:]]
    values_h = dict()
    tolerance = 1e-2
    kde = KDE(0)
    kde.fit(X_train)
    def get_value(h):
        if h in values_h.keys():
            return values_h[h]
        kde.h = np.exp(h)
        value = kde.loglikelihood(X_validation)
        values_h[h] = value
        return value
    
    test_h = np.log(constant_h(X_train))
    if get_value(test_h) < get_value(test_h - 1):
        change = -1
    else:
        change = 1
    while get_value(test_h) < get_value(test_h + change):
        test_h = test_h + change
        change = 2 * change
        
    
    best_h = max(values_h, key=values_h.get)
    list_h = np.array(list(values_h.keys()))
    lower, middle, upper = max(list_h[list_h < best_h]), best_h, min(list_h[list_h > best_h])
    while lower + tolerance < upper:
        coeffs = lower, middle, upper
        values = [get_value(coeff) for coeff in coeffs]
        coeffs = [*coeffs, *coeffs]
        new_h = \
            sum([(coeffs[i+1]**2 - coeffs[i+2]**2) * values[i] for i in range(3)]) /\
            sum([(coeffs[i+1] - coeffs[i+2]) * values[i] for i in range(3)]) / 2
        if abs(new_h - middle) < tolerance:
            if upper - middle < middle - lower:
                new_h = middle - tolerance / 3
            else:
                new_h = middle + tolerance / 3
        if get_value(middle) < get_value(new_h):
            if middle < new_h:
                lower, middle, upper = middle, new_h, upper
            else:
                lower, middle, upper = lower, new_h, middle
        else:
            if middle < new_h:
                lower, middle, upper = lower, middle, new_h
            else:
                lower, middle, upper = new_h, middle, upper
    
    return np.exp(middle)

class KDE(Generator):
    def __init__(self, h: 'float | str | function' = 'tune'):
        super().__init__()
        if isinstance(h, str):
            if h.lower() in ['auto', 'tune']:
                self.h_generator = optimize_h
            elif h.lower() in ['constant', 'c']:
                self.h_generator = constant_h
            else:
                raise ValueError(f"Unrecognized h type ({h})")
        elif isinstance(h, int) or isinstance(h, float):
            self.h_generator = lambda X: h
        else:
            self.h_generator = h
    
    @fit_method
    def fit(self, X):
        self.X = X
        self.n, self.dim = X.shape
        self.h = self.h_generator(X)
       
    @if_fitted 
    def probabilities(self, X):
        assert X.shape[1] == self.dim, "Size mismatch"
        dist = np.sum(np.square(
            np.reshape(self.X, (1, -1, self.dim))
            - np.reshape(X, (-1, 1, self.dim))
            ), 2)
        probs = np.sum(np.exp(-dist / 2 / self.h ** 2) / np.power(2 * np.pi * self.h ** 2, self.dim / 2), 1) / self.n
        return probs
      
    @if_fitted
    def generate(self, size):
        ind = np.random.choice(self.n, size)
        S = np.random.normal(self.X[ind], self.h)
        return S
    
    @if_fitted
    def fill(self, Y):
        assert Y.shape[1] == self.dim, "Size mismatch"
        diffs = np.reshape(self.X, (1, -1, self.dim)) - np.reshape(Y, (-1, 1, self.dim))
        dists = np.sum(np.square(np.nan_to_num(diffs)), 2)
        probs = np.exp(-dists / 2 / self.h ** 2)
        for y, prob in zip(Y, probs):
            bad = np.isnan(y)
            ind = np.random.choice(self.n, p=prob / np.sum(prob))
            y[bad] = np.random.normal(self.X[ind, bad], self.h)
        return Y
    


# -*- coding: utf-8 -*-
"""
Gaussian Mixture Model

@author: david
"""

import numpy as np
from time import time as gettime

from ._base import Generator, if_fitted, fit_method

PHI = (np.sqrt(5) - 1) / 2

class Bayesian():
    def __init__(self, X):
        self.X = X
        self.n, self.dim = X.shape
        
    def data(self):
        return self.X
    
    def value(self, cls):
        return cls._degrees_freedom() * np.log(self.n) - 2 * cls.loglikelihood(self.X)

class Akaike():
    def __init__(self, X):
        self.X = X
        self.n, self.dim = X.shape
        
    def data(self):
        return self.X
    
    def value(self, cls):
        return 2 * cls._degrees_freedom() - 2 * cls.loglikelihood(self.X)

class CrossValidation():
    def __init__(self, X):
        self.X = X
        self.n, self.dim = X.shape
        self.ind = np.random.choice(self.n, self.n, False)
        self.separator = int(self.n * 0.6)
        
    def data(self):
        return self.X[self.ind[:self.separator]]
    
    def value(self, cls):
        return - cls.loglikelihood(self.X[self.ind[self.separator:]])

def parse_criterion(criterion, X):
    if not isinstance(criterion, str):
        raise ValueError('Criterion must be a string')
    elif criterion.lower() in ['bayesian', 'b', 'bic']:
        return Bayesian(X)
    elif criterion.lower() in ['akaike', 'a', 'aic']:
        return Akaike(X)
    elif criterion.lower() in ['cross-validation', 'cv']:
        return CrossValidation(X)
    raise ValueError(f'{criterion} is not a recognized criterion, use AIC, CV or BIC instead')

class GMM(Generator):
    def __init__(self, k: 'int | None' = None, k_max: 'int | None' = None, multivariate: 'bool' = True,
                 iterations_limit: 'int' = 1000, time_limit: 'float' = 1,
                 llh_tolerance: 'float' = 1e-3, attempts: 'int' = 3, criterion: 'str' = 'BIC'):
        super().__init__()
        if k is None:
            self.k_mode = 'search'
            self.k_max = k_max
            self.k = None
        elif isinstance(k, int):
            self.k_mode = 'set'
            self.k_max = None
            self.k = max(k, 1)
        else:
            raise ValueError("k must be an integer or None")
        self.multivariate = multivariate
        self.iterations_limit = iterations_limit
        self.time_limit = time_limit
        self.llh_tolerance = llh_tolerance
        self.attempts = attempts
        self.criterion = criterion
    
    @fit_method
    def fit(self, X):
        self.X = X
        self.n, self.dim = X.shape
        if self.k_mode == 'search':
            # Search optimal K value
            criterion = parse_criterion(self.criterion, X)
            X_search = criterion.data()
            gmm = GMM(
                multivariate=self.multivariate, \
                iterations_limit=self.iterations_limit, \
                time_limit=self.time_limit, \
                llh_tolerance=self.llh_tolerance, \
                attempts=self.attempts
            )
            if self.k_max is None:
                criterion_values = dict()
                best_value = np.Inf
                k = 1
                loop = True
                while loop:
                    loop = False
                    k = 2 * k
                    gmm._set_k(k)
                    gmm.fit(X_search)
                    criterion_values[k] = criterion.value(gmm)
                    if criterion_values[k] < best_value:
                        best_value = criterion_values[k]
                        best_k = k
                        loop = True
                best_k = gmm._optimize_in(X_search, criterion, \
                                          precision=int(np.sqrt(best_k)), \
                                          min_k=int(best_k / 2), max_k=int(best_k * 2), \
                                          criterion_values=criterion_values)
            else:
                best_k = gmm._optimize_in(X_search, criterion, \
                                          min_k=1, max_k=self.k_max)
            self.k = best_k
          
        # Run Expectation-Maximization algorithm
        best_llh = None
        for attempt in range(self.attempts):
            self._reset()
            loop = True
            iteration_left = self.iterations_limit
            time_limit = self.time_limit + gettime()
            llh = self.loglikelihood(X)
            while loop:
                self._iterate(self.X)
                self._step()
                    
                iteration_left -= 1
                last_llh = llh
                llh = self.loglikelihood(X)
                
                loop &= iteration_left > 0
                loop &= llh - last_llh > self.llh_tolerance
                loop &= time_limit > gettime()
                
            if not best_llh or llh > best_llh:
                best_llh = llh
                best_params = (self.weights, self.means, self.covariances, self.inverses, self.determinants)
        self.weights, self.means, self.covariances, self.inverses, self.determinants = best_params
    
    @if_fitted
    def probabilities(self, X):
        return np.sum(self._k_probability(X), 1)
    
    @if_fitted   
    def generate(self, size):
        ind, num = np.unique(np.random.choice(self.k, size, p=self.weights), return_counts=True)
        s = [
            np.random.multivariate_normal(self.means[i], self.covariances[i], n)
            for i, n in zip(ind, num)
            ]
        S = np.concatenate(s)
        return S
    
    @if_fitted
    def fill(self, Y):
        assert Y.shape[1] == self.dim, "Size mismatch"
        for y in Y:
            bad = np.isnan(y)
            good = np.logical_not(bad)
            goods = np.sum(good)
            k_prob = []
            tol = 1e-6
            for weight, mean, covariance in \
                zip(self.weights, self.means, self.covariances):
                    centered = (y - mean)[good]
                    _covariance = covariance[good][:,good] + tol * np.identity(goods)
                    determinant = np.linalg.det(_covariance)
                    inverse = np.linalg.inv(_covariance)
                    exponent = (centered @ inverse @ centered) / 2
                    k_prob.append(weight * np.exp(-exponent) / np.sqrt(determinant))
            prob = np.array(k_prob)
            ind = np.random.choice(self.k, p=prob / np.sum(prob))
            covariance = self.covariances[ind]
            inv_subcov = np.linalg.inv(covariance[good][:,good] + tol * np.identity(goods))
            new_mean = self.means[ind][bad] + covariance[bad][:,good] @ inv_subcov @ (y - self.means[ind])[good]
            new_cov = covariance[bad][:,bad] - covariance[bad][:,good] @ inv_subcov @ covariance[good][:,bad]
            y[bad] = np.random.multivariate_normal(new_mean, new_cov)
        return Y
        
    def _reset(self):
        self.weights = np.zeros(self.k) + 1 / self.k
        self.means = np.random.normal(size=(self.k, self.dim))
        self.covariances = np.array([np.identity(self.dim) for _ in range(self.k)])
        self.inverses = np.array([np.identity(self.dim) for _ in range(self.k)])
        self.determinants = np.ones(self.k)
        
    def _set_k(self, k):
        self.k_mode = 'set'
        self.k_max = None
        self.k = max(k, 1)
    
    def _k_probability(self, X):
        assert X.shape[1] == self.dim, "Size mismatch"
        k_prob = []
        for weight, mean, inverse, determinant in \
            zip(self.weights, self.means, self.inverses, self.determinants):
                centered = X - mean
                exponent = -np.sum(centered.dot(inverse) * centered, 1) / 2
                k_prob.append(weight * np.exp(exponent) / np.sqrt(determinant))
                    
        k_prob = np.array(k_prob).transpose() / np.power(2 * np.pi, self.dim / 2)
        return np.nan_to_num(k_prob)
    
    def _iterate(self, X):
        # Expectation
        probs = self._k_probability(X).transpose()
        resps = probs / np.maximum(np.sum(probs, 0), 1e-300)
        t_resps = np.maximum(np.sum(resps, 1), 1e-300)
        
        # Maximization
        self.new_weights = t_resps / self.n
        self.new_weights /= np.sum(self.new_weights)
        self.new_means = np.transpose(resps.dot(X).T / t_resps)
        self.new_covariances = np.array([
            ((X - mean).T * resp) @ (X - mean) / t_resp 
            for t_resp, resp, mean in 
                zip(t_resps, resps, self.new_means)
                ])
    
    def _step(self):
        self.weights = self.new_weights
        self.means = self.new_means
        diag = np.arange(self.dim)
        if self.multivariate:
            self.covariances = self.new_covariances
            self.covariances[:, diag, diag] += 1e-6
        else:
            self.covariances[:, diag, diag] = self.new_covariances[:, diag, diag] + 1e-6
        for i in range(self.k):
            if self.multivariate:
                self.inverses[i] = np.linalg.inv(self.covariances[i])
                self.determinants[i] = np.linalg.det(self.covariances[i])
            else:
                self.inverses[i, diag, diag] = 1 / self.covariances[i].diagonal()
                self.determinants[i] = np.prod(self.covariances[i].diagonal())
    
    def _degrees_freedom(self):
        if self.multivariate:
            return self.k * (self.dim ** 2 + 3 * self.dim + 2) / 2 - 1
        else:
            return self.k * (2 * self.dim + 1) - 1
    
    def _optimize_in(self, X, criterion, max_k, min_k=1, precision=1, criterion_values = dict()):
        def get_value(k):
            if k in criterion_values.keys():
                return criterion_values[k]
            self._set_k(k)
            self.fit(X)
            value = criterion.value(self)
            criterion_values[k] = value
            return value
        
        low, high, size = min_k, max_k, max_k - min_k
        phi_low, phi_high = round(high - size * PHI), round(low + size * PHI)
        while phi_high >= phi_low + precision:
            if get_value(phi_high) < get_value(phi_low):
                low = phi_low
            else:
                high = phi_high
            size = high - low
            phi_low, phi_high = round(high - size * PHI), round(low + size * PHI)
        get_value(low), get_value(phi_low), get_value(high)
        best_k = min(criterion_values, key=criterion_values.get)
        self._set_k(best_k)
        return best_k
            



# -*- coding: utf-8 -*-
"""
Variational AutoEncoder

@author: david
"""

import numpy as np
import torch
import optuna
from torch import nn
from torch.optim.lr_scheduler import ReduceLROnPlateau
from time import time as gettime

from ._base import Generator, if_fitted, fit_method

def loss_function(recon_x, x, mu, logvar):
    recLoss = nn.MSELoss(reduction='sum')
    KLD = 0.5 * torch.sum(logvar.exp() + mu.pow(2) - 1 - logvar)
    return recLoss(recon_x, x) + torch.sqrt(KLD)

class VAE(Generator, nn.Module):
    def __init__(self, device: 'str | device' = 'auto',
                 layers_sizes: 'list | None' = None, 
                 latent_dimension: 'int | None' = None, layers: 'int | None' = None,
                 epochs: 'int | None' = None, search_trials: 'int' = 10,
                 batches: 'int | None' = None, batch_size: 'int | None' = 128,
                 learning_rate: 'float' = 1e-1,
                 reduce_learning_rate: 'bool' = True, min_learning_rate: 'float' = 1e-5,
                 patience: 'int | None' = None, time_limit: 'float' = 10):
        super().__init__()
        if device == 'auto':
            self._device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        elif type(device) == str:
            self._device = torch.device(device)
        else:
            self._device = device
        if layers_sizes is None:
            self.layers_sizes = None
            self.latent_dimension = latent_dimension
            self.layers = layers
        else:
            self.layers_sizes = layers_sizes
            self.latent_dimension = layers_sizes[-1]
            self.layers = len(layers_sizes)
        self.epochs = epochs
        self.search_trials = search_trials
        self.batches = batches
        self.batch_size = batch_size
        self.learning_rate = learning_rate
        self.reduce_learning_rate = reduce_learning_rate
        self.min_learning_rate = min_learning_rate
        self.patience = patience
        self.time_limit = time_limit
    
    @fit_method
    def fit(self, X):
        torch_X = torch.from_numpy(X).float().to(self._device)
        self.n, self.dim = X.shape
        
        if self.layers_sizes is None:
            params = {
                'latent_dimension': self.latent_dimension,
                'layers': self.layers
            }
            
            if self.latent_dimension is None or self.layers is None:
                optuna.logging.set_verbosity(optuna.logging.ERROR)
                study = optuna.create_study()
                study.optimize(self._get_objective(torch_X), n_trials=self.search_trials)
                params = params | study.best_params
            latent_dimension = params['latent_dimension']
            layers = params['layers']
            
            layers_sizes = np.array(np.round(np.exp(np.linspace(np.log(self.dim), np.log(latent_dimension), layers + 1))), dtype=int)
        else:
            layers_sizes = [self.dim] + self.layers_sizes
        
        return self._run_configuration(torch_X, layers_sizes)
        
    @if_fitted
    def generate(self, size):
        with torch.no_grad():
            S = torch.zeros((size, self.used_latent_dimension), device=self._device, dtype=torch.float).normal_()
            X = self.decode(S)
        return X.detach().cpu().numpy().astype(float)
    
    # VAE forward steps
    def encode(self, x):
        enc = self.encoder(x)
        return self.mean(enc), self.logvar(enc)

    def reparametrize(self, mean, logvar):
        std = torch.exp(0.5 * logvar)
        eps = torch.zeros(std.size(), device=self._device.type, dtype=torch.float).normal_()
        return mean + eps * std

    def decode(self, z):
        dec = self.decoder(z)
        return dec

    def forward(self, x):
        mean, logvar = self.encode(x)
        z = self.reparametrize(mean, logvar)
        return self.decode(z), mean, logvar
    
    # VAE training
    def _train(self, X):
        time_limit = self.time_limit + gettime()
        batches = max(round(self.n / self.batch_size), 1) if self.batches is None else self.batches
        batch_size = int(np.ceil(self.n / batches))
        train_loader = torch.utils.data.DataLoader(X, batch_size=batch_size, shuffle=True)
        optimizer = torch.optim.Adam(self.parameters(), lr=self.learning_rate)
        
        loop = True
        if self.epochs is not None:
            mode = 'epochs'
            current_epoch = 0
            if self.reduce_learning_rate:
                patience = 10 if self.patience is None else self.patience
                schedule = ReduceLROnPlateau(optimizer, patience=patience)
        elif self.reduce_learning_rate:
            mode = 'lr'
            patience = 5 if self.patience is None else self.patience
            schedule = ReduceLROnPlateau(optimizer, patience=patience)
        else:
            mode = 'patience'
            patience = 20 if self.patience is None else self.patience
            best_loss = None
            since_improvement = 0
            
        self.train()
        while loop:
            loss = 0
            for batch in train_loader:
                batch = batch.to(self._device)
                optimizer.zero_grad()
                
                outputs, mean, logvar = self.forward(batch)
                
                train_loss = loss_function(outputs, batch, mean, logvar)
                train_loss.backward()
                
                optimizer.step()
                
                loss += train_loss.detach().cpu()
            if self.reduce_learning_rate:
                schedule.step(loss)
                
            
            if mode == 'epochs':
                current_epoch += 1
                loop = current_epoch < self.epochs
            elif mode == 'lr':
                loop = optimizer.state_dict()['param_groups'][0]['lr'] > self.min_learning_rate
            else:
                since_improvement += 1
                if best_loss is None or best_loss > loss:
                    best_loss = loss
                    since_improvement = 0
                loop = since_improvement < patience
            loop &= time_limit > gettime()
        self.eval()
    
    def _run_configuration(self, X, layers_sizes):
        latent_dimension = layers_sizes[-1]
        self.used_latent_dimension = latent_dimension
        
        f = nn.Sigmoid()
        
        self.encoder = nn.Sequential(*[
            f if i == -1 else nn.Linear(layers_sizes[i], layers_sizes[i + 1])
            for i in [e // 2 if e % 2 == 0 else -1 for e in range(2 * len(layers_sizes) - 4)]
            ])
        self.mean = nn.Linear(layers_sizes[-2], latent_dimension)
        self.logvar = nn.Linear(layers_sizes[-2], latent_dimension)
        self.decoder = nn.Sequential(*[
            f if i == -1 else nn.Linear(layers_sizes[i], layers_sizes[i - 1])
            for i in [e // 2 if e % 2 == 0 else -1 for e in range(2 * len(layers_sizes) - 2, 1, -1)]
            ])
        
        self.to(self._device)
        self._train(X)
        with torch.no_grad():
            recLoss = nn.MSELoss(reduction='mean')
            mean, logvar = self.encode(X)
            Y = self.decode(mean)
            loss = recLoss(X, Y)
            # noise lost because of dimensionality reduction
            # Noise = {Uncorrelated variance} * {Average distance to 0 of N(0, 1)} * {#Lost dimensions}
            noise = 0.05 * np.sqrt(2 / np.pi) * (self.dim - latent_dimension)
        return loss - noise
    
    def _get_objective(self, X):
        def objective(trial):
            latent_dimension = self.latent_dimension
            layers = self.layers
            latent_dimension = trial.suggest_int('latent_dimension', 1, self.dim) \
                            if latent_dimension is None \
                            else latent_dimension
            layers = trial.suggest_int('layers', 1, 5) if layers is None else layers
            
            layers_sizes = np.array(np.round(np.exp(np.linspace(np.log(self.dim), np.log(latent_dimension), layers + 1))), dtype=int)
            
            return self._run_configuration(X, layers_sizes)
        return objective
    
    def prefit(self, X):
        self.n, self.dim = X.shape
        torch_X = torch.from_numpy(X).float().to(self._device)
        
        optuna.logging.set_verbosity(optuna.logging.ERROR)
        study = optuna.create_study()
        study.optimize(self._get_objective(torch_X), n_trials=self.search_trials)
        
        params = {
            'latent_dimension': self.latent_dimension,
            'layers': self.layers
        } | study.best_params
        self.latent_dimension = params['latent_dimension']
        self.layers = params['layers']
        
        # Clean neural network
        self.encoder = None
        self.mean = None
        self.logvar = None
        self.decoder = None


# -*- coding: utf-8 -*-
"""
Base Generator

@author: david
"""

import numpy as np
import functools

def if_fitted(func):
    @functools.wraps(func)
    def wrapper(self, *args, **kwargs):
        assert self.fitted
        return func(self, *args, **kwargs)
    return wrapper

def fit_method(func):
    @functools.wraps(func)
    def wrapper(self, *args, **kwargs):
        self.fitted = True
        return func(self, *args, **kwargs)
    return wrapper

class Generator():
    def __init__(self):
        super().__init__()
        self.fitted = False
       
    @if_fitted
    def loglikelihood(self, X):
        return np.sum(np.log(self.probabilities(X) + 1e-300))
    


from ._base import Generator, if_fitted, fit_method
from ._GMM import GMM
from ._KDE import KDE

class AutoSelection(Generator):
    def __init__(self):
        super().__init__()
        self.generator = None

    @fit_method
    def fit(self, X):
        self.X = X
        self.n, self.dim = X.shape
        if 200 <= self.n <= 2e3:
            self.generator = GMM()
        else:
            self.generator = KDE()
        self.generator.fit(X)
       
    @if_fitted 
    def probabilities(self, X):
        return self.generator.probabilities(X)
      
    @if_fitted
    def generate(self, size):
        return self.generator.generate(size)
    
    @if_fitted
    def fill(self, Y):
        return self.generator.fill(Y)


# -*- coding: utf-8 -*-
"""
Created on Mon Oct 10 18:37:51 2022

@author: david
"""

import numpy as np

from ._classes import EncoderNone, EncoderDiscrete, EncoderIgnore, EncoderOHE, EncoderLimit, EncoderScale
    
def greater(value: 'float' = 0, include: 'bool' = True, influence: 'float' = 1):
    """
    Encoder for values greater than value.

    Parameters
    ----------
    value : 'float', optional
        Minimum value. The default is 0.
    include : 'bool', optional
        Include minimum in posible values. The default is True.
    influence : 'float', optional
        Range of improbable values near the minimum when not included. The default is 1.

    Returns
    -------
    Encoder.
    
    """
    if include:
        return EncoderLimit(lower=value, tails=False)
    else:
        return EncoderLimit(lower=value, tails=True, influence=influence)
    
def lower(value: 'float' = 0, include: 'bool' = True, influence: 'float' = 1):
    """
    Encoder for values greater than value.

    Parameters
    ----------
    value : 'float', optional
        Minimum value. The default is 0.
    include : 'bool', optional
        Include minimum in posible values. The default is True.
    influence : 'float', optional
        Range of improbable values near the minimum when not included. The default is 1.

    Returns
    -------
    Encoder.
    
    """
    if include:
        return EncoderLimit(upper=value, tails=False)
    else:
        return EncoderLimit(upper=value, tails=True, influence=influence)

def auto(data: 'DataFrame'):
    """
    Select Encoder based on dataframe column.

    Parameters
    ----------
    data : 'DataFrame'
        Data frame column.

    Returns
    -------
    Encoder.

    """
    if data.dtype == object:
        symbols = data.unique()
        n = len(data)
        s = len(symbols)
        if s == 1:
            return EncoderIgnore(symbols[0])
        elif s == 2:
            return EncoderScale(symbols)
        elif np.sqrt(n) > s and s < 100:
            return EncoderOHE(list(symbols))
        else:
            return EncoderIgnore('Ignored')
    else:
        if data.dtype == np.float64:
            if (data > 1e-6).all():
                return EncoderLimit(lower=0, tails=True)
            elif (data >= -0).all():
                return EncoderLimit(lower=0, tails=False)
            else:
                return EncoderNone()
        else:
            minimum = None
            if (data > 0).all():
                minimum = 1
            elif (data >= 0).all():
                minimum = 0
            return EncoderDiscrete(minimum=minimum)